/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/


#include <daemon/backend/data/chunk_storage.hpp>
#include <daemon/backend/metadata/db.hpp>
#include <daemon/daemon.hpp>
#include <daemon/handler/rpc_defs.hpp>
#include <daemon/handler/rpc_util.hpp>
#include <daemon/relocation/transmitter.hpp>
#include <global/rpc/rpc_types.hpp>

#include <iostream>

#include <fmt/format.h>

using namespace std;

/*
 * This file contains Margo RPC handlers to receive metadata and data from
 * other daemons during relocation as well as a handler for relocation
 * management.
 */

namespace {

/**
 * Receive and store a metadata entry.
 */
hg_return_t
rpc_srv_relocate_metadata(hg_handle_t handle) {
    rpc_relocate_metadata_in_t in{};
    rpc_err_out_t out{};
    // out.err = EIO;

    // Get input parmeters
    auto ret = margo_get_input(handle, &in);
    if(ret != HG_SUCCESS) {
        GKFS_DATA->spdlogger()->error(
                "{}() Could not get RPC input data with err '{}'", __func__,
                ret);
        out.err = EBUSY;
        return gkfs::rpc::cleanup_respond(&handle, &in, &out);
    }

    cout << fmt::format("TODO(dauer) Received metadata entry {}: {}\n", in.key,
                        in.value);

    GKFS_DATA->mdb()->put(in.key, in.value);

    out.err = 0; // TODO(dauer)
    return gkfs::rpc::cleanup_respond(&handle, &in, &out);
}

/**
 * Receive and store a chunk.
 */
hg_return_t
rpc_srv_relocate_chunk(hg_handle_t handle) {
    rpc_relocate_chunk_in_t in{};
    rpc_err_out_t out{};
    // out.err = EIO;

    // Get input parmeters
    auto ret = margo_get_input(handle, &in);
    if(ret != HG_SUCCESS) {
        GKFS_DATA->spdlogger()->error(
                "{}() Could not get RPC input data with err '{}'", __func__,
                ret);
        out.err = EBUSY;
        return gkfs::rpc::cleanup_respond(&handle, &in, &out);
    }
    auto hgi = margo_get_info(handle);
    auto mid = margo_hg_info_get_instance(hgi);
    auto bulk_size = margo_bulk_get_size(in.bulk_handle);

    cout << fmt::format("TODO(dauer) Received chunk {}: chunk {} size {}\n",
                        in.path, in.chunk_id, bulk_size);

    unique_ptr<char[]> buf(new char[bulk_size]());
    char* buf_ptr = buf.get();
    hg_bulk_t local_bulk{};
    ret = margo_bulk_create(mid, 1, (void**) &buf_ptr, &bulk_size,
                            HG_BULK_READWRITE, &local_bulk);
    assert(ret == HG_SUCCESS);

    ret = margo_bulk_transfer(mid, HG_BULK_PULL, hgi->addr, in.bulk_handle, 0,
                              local_bulk, 0, bulk_size);
    assert(ret == HG_SUCCESS);


    // write data (blocking)
    GKFS_DATA->storage()->write_chunk(in.path, in.chunk_id, buf.get(),
                                      bulk_size, 0);

    out.err = 0; // TODO(dauer)
    return gkfs::rpc::cleanup_respond(&handle, &in, &out, &local_bulk);
}

/**
 * Receive the signal to start relocation.
 */
hg_return_t
rpc_srv_relocation_start(hg_handle_t handle) {
    cout << "TODO(dauer) relocation start received\n";
    rpc_relocation_start_in_t in{};
    rpc_err_out_t out{};
    out.err = EIO;
    // Getting some information from margo
    auto ret = margo_get_input(handle, &in);
    if(ret != HG_SUCCESS) {
        GKFS_DATA->spdlogger()->error(
                "{}() Could not get RPC input data with err {}", __func__, ret);
        out.err = EBUSY;
        return gkfs::rpc::cleanup_respond(&handle, &in, &out);
    }

    gkfs::relocation::transmit_metadata_and_data(in.host_id);

    out.err = 0; // TODO(dauer)
    return gkfs::rpc::cleanup_respond(&handle, &in, &out);
}

} // namespace

DEFINE_MARGO_RPC_HANDLER(rpc_srv_relocate_metadata)

DEFINE_MARGO_RPC_HANDLER(rpc_srv_relocate_chunk)

DEFINE_MARGO_RPC_HANDLER(rpc_srv_relocation_start)
