/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/


#include <daemon/relocation/config_manager.hpp>
#include <daemon/daemon.hpp>
#include <global/global_defs.hpp>
#include <daemon/handler/rpc_defs.hpp>
#include <daemon/handler/rpc_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/random_slicing/DistRandSlice.hpp>
#include <global/rpc/rpc_types.hpp>
#include <global/rpc/distributor.hpp>

#include <algorithm>
#include <cassert>
#include <filesystem>
#include <iostream>
#include <list>
#include <set>

#include <fmt/format.h>

// margo
extern "C" {
#include <margo.h>
}

using namespace std;

namespace gkfs::relocation {

void
calculate_random_slicing() {
    using gkfs::rpc::host_t;

    // use id only for new configurations!
    auto hostsfile_hosts = read_hosts_file();
    list<VDRIVE::Disk*> disks;
    auto filename = GKFS_DATA->hosts_file() + ".rsconfig";
    auto dist = std::make_shared<VDRIVE::DistRandSlice>();
    if(std::filesystem::exists(filename)) {
        GKFS_DATA->spdlogger()->info(
                "Existing random slicing configuration found, "
                "loading from file: {}",
                filename);
        dist->from_file(filename);

        // update configuration

        // create distributor
        gkfs::rpc::RandomSlicingDistributor rsdist(dist);

        // get vector of all existing rs hosts
        auto rs_hosts = rsdist.get_hosts_map();
        uint64_t max_host_id = 0;
        std::set<std::string> rs_uris{};
        for(const auto& [id, host] : rs_hosts) {
            max_host_id = std::max(max_host_id, id);
            rs_uris.insert(host.second);
        }

        std::vector<std::string> gone_uris{};
        std::vector<uint64_t> gone_ids{};
        std::map<std::string, std::string> new_uris_hostname{};
        std::set<std::string> hostsfile_uris{};

        std::string new_uris_str{};
        std::string gone_uris_str{};

        for(const auto& [dont_care, host] : hostsfile_hosts) {
            const auto& [hostname, uri] = host;
            hostsfile_uris.insert(uri);
            if(rs_uris.count(uri) == 0) {
                new_uris_hostname[uri] = hostname;
                new_uris_str += "\n" + uri;
            }
        }

        for(const auto& [id, host] : rs_hosts) {
            const auto& [hostname, uri] = host;
            if(hostsfile_uris.count(uri) == 0) {
                gone_uris.push_back(uri);
                gone_ids.push_back(id);
                gone_uris_str += "\n" + uri;
            }
        }


        GKFS_DATA->spdlogger()->info("Hosts to be added: {}", new_uris_str);
        GKFS_DATA->spdlogger()->info("Hosts to be removed: {}", gone_uris_str);

        /*
                // gone = rs_hosts \ hostsfile_hosts
                std::set_difference(all_rs_hosts.begin(), all_rs_hosts.end(),
                                    all_hostsfile_hosts.begin(),
                                    all_hostsfile_hosts.end(),
                                    std::inserter(gone_uris,
           gone_uris.begin()));

                // new = hostsfile_hosts \ rs_hosts
                std::set_difference(all_hostsfile_hosts.begin(),
                                    all_hostsfile_hosts.end(),
           all_rs_hosts.begin(), all_rs_hosts.end(), std::inserter(new_uris,
           new_uris.begin()));
        */

        // remove gone hosts
        if(gone_ids.size() > 0) {
            dist->removeDisks(gone_ids);
        }
        if(new_uris_hostname.size() > 0) {
            // add new hosts
            for(const auto& [uri, hostname] : new_uris_hostname) {
                disks.emplace_back(
                        new VDRIVE::Disk(++max_host_id, 1, hostname, uri));
            }
            dist->addDisks(&disks);
        }
    } else {
        GKFS_DATA->spdlogger()->info(
                "No existing random slicing configuration found, "
                "creating from scratch with {} hosts.",
                hostsfile_hosts.size());

        for(const auto& [id, host] : hostsfile_hosts) {
            const auto& [hostname, uri] = host;
            disks.emplace_back(new VDRIVE::Disk(id, 1, hostname, uri));
            cout << uri << "\n";
        }
        dist->setConfiguration(&disks, 1, 1);
    }
    for(auto disk : disks) {
        delete disk;
    }

    dist->to_file(filename);

    // Just test reading here
    VDRIVE::DistRandSlice dist_read{};
    dist_read.from_file(filename);
    dist_read.to_file(filename + ".debug-reexport");
}

unordered_map<uint64_t, pair<string, string>>
read_hosts_file() {
    string hostfile = GKFS_DATA->hosts_file();

    unordered_map<uint64_t, pair<string, string>> hosts;
    try {
        hosts = gkfs::utils::load_hostfile(hostfile);
    } catch(const exception& e) {
        auto emsg = fmt::format("Failed to load hosts file: {}", e.what());
        throw runtime_error(emsg);
    }

    if(hosts.empty()) {
        throw runtime_error(fmt::format("Hostfile empty: '{}'", hostfile));
    }

    GKFS_DATA->spdlogger()->info("Hosts pool size: {}", hosts.size());

    return hosts;
}

/**
 * Send RPCs to all daemons in hosts file to trigger relocation.
 */
void
do_relocation() {
    auto hosts = read_hosts_file();

    cout << "number of hosts: " << hosts.size() << "\n";
    auto mid =
            margo_init(GKFS_DATA->bind_addr().c_str(), MARGO_CLIENT_MODE, 0, 0);
    if(mid == MARGO_INSTANCE_NULL) {
        throw runtime_error("Failed to initialize the Margo RPC client");
    }

    hg_id_t relocation_start_id = MARGO_REGISTER(
            mid, gkfs::rpc::tag::relocation_start, rpc_relocation_start_in_t,
            rpc_err_out_t, rpc_srv_relocation_start);


    for(const auto& [host_id, host] : hosts) {
        const auto& [hostname, uri] = host;
        cout << fmt::format("Hostname: {} Address: {} ID: {}\n", hostname, uri,
                            host_id);
        hg_addr_t host_addr;
        rpc_relocation_start_in_t in{};
        rpc_err_out_t out{};
        in.host_id = host_id;
        auto ret = margo_addr_lookup(mid, uri.c_str(), &host_addr);
        assert(ret == HG_SUCCESS);

        // let's do this sequential first
        hg_handle_t handle;
        ret = margo_create(mid, host_addr, relocation_start_id, &handle);
        assert(ret == HG_SUCCESS);

        ret = margo_forward(handle, &in); // blocking
        assert(ret == HG_SUCCESS);

        ret = margo_get_output(handle, &out);
        assert(ret == HG_SUCCESS);

        // TODO(dauer) process output

        if(HG_SUCCESS !=
           gkfs::rpc::margo_client_cleanup(&handle, &out, &mid, &host_addr)) {
            cout << "Error during margo cleanup.\n";
        }
    }
    margo_finalize(mid);
}

} // namespace gkfs::relocation